-- TODO: Make this a true module, by returning it inside a local table...

------------------------------------------------------
-----------------  XML Helper Module   ------------------
------------------------------------------------------

-- Externs:
-- XMLToTable(strXML))
-- XMLPath(tXML, strWalk)
-- TableToXML(tXML)
  -- Examples:
  -- {params = {param = { name = "MyValue", value = {static = {type = {INTEGER = 123}}}}}}
  -- {http_api = {session = {device_id = 1, tpnumber = 3}}}
-- MakeXML(strTag, strValue, tAttrs)
  -- Examples:
  -- MakeXML("c4soap", table.concat(params, ''), {name = strCommand, seq = mSeq, async = TF[bAsync]})
-- XMLSpecialChars

do
  XMLHelper = {}

  local XMLCHARS = {
    ["&apos;"] = "'",
    ["&lt;"] = "<",
    ["&gt;"] = ">",
    ["&quot;"] = "'",
    ["&amp;"] = "&",
  }


  function XMLHelper.XMLSpecialChars(name)
    local s_gsub = string.gsub
    for k,v in pairs(XMLCHARS) do name = s_gsub(name, k, v) end
    return name
  end


  ---- SIMPLE XML PARSER from http://lua-users.org/wiki/LuaXml (parseargs and collect originally)------------
  local function parseargs(s)
    local arg = {}
    string.gsub(s, "([%w_]+)=([\"'])(.-)%2", function (w, _, a)
      arg[w] = a
    end)
    return arg
  end

  function XMLHelper.XMLToTable(s)
    local stack = {}
    local top = {}
    table.insert(stack, top)
    local ni,c,label,xarg, empty
    local i, j = 1, 1
    while true do
      ni,j,c,label,xarg, empty = string.find(s, "<(%/?)([%w_]+)(.-)(%/?)>", i)
      if not ni then break end
      local text = string.sub(s, i, ni-1)
      if not string.find(text, "^%s*$") then
        table.insert(top, text)
      end
      if empty == "/" then  -- empty element tag
        table.insert(top, {label=label, xarg=parseargs(xarg), empty=1})
      elseif c == "" then   -- start tag
        top = {label=label, xarg=parseargs(xarg)}
        table.insert(stack, top)   -- new level
      else  -- end tag
        local toclose = table.remove(stack)  -- remove top
        top = stack[#stack]
        if #stack < 1 then
          error("nothing to close with "..label)
        end
        if toclose.label ~= label then
          error("trying to close "..toclose.label.." with "..label)
        end
        table.insert(top, toclose)
      end
      i = j+1
    end
    local text = string.sub(s, i)
    if not string.find(text, "^%s*$") then
      table.insert(stack[#stack], text)
    end
    if #stack > 1 then
      error("unclosed "..stack[#stack].label)
    end
    return stack[1]
  end


  -- Converts tables with sub-tables into <key><subkey>subvalue</subkey></key> XML...
  -- NOTE: Does *not* support attributes...
  function XMLHelper.TableToXML(tXML)
    if (type(tXML) ~= "table") then return tostring(tXML) end
    local val = ""

    local k,v = next(tXML)
    while (k ~= nil) do
      val = val .. XMLHelper.MakeXML(k, XMLHelper.TableToXML(v))
      k,v = next(tXML, k)
    end
    return val
  end


  function XMLHelper.MakeXML(strTag, strValue, tAttrs)
    local xml, p = {}, {}
    for k,v in pairs(tAttrs or {}) do
      table.insert(p, ' ' .. k .. '="' .. v .. '"')
    end
    table.insert(xml, '<' .. strTag .. table.concat(p, '') .. '>' .. strValue .. '</' .. strTag .. '>')
    return table.concat(xml, '')
  end


  function XMLHelper.XMLPath(tXML, strWalk)
    local out = {}
    -- Do a sort-of-like XPath walk through the Tableified XML...
    -- Examples:
    -- XMLHelper.XMLPath(t, "^c4soap")  -- returns inner *text* of <c4soap attr="Value1">InnerText</c4soap>
    -- XMLHelper.XMLPath(t, "c4soap/physicaldevices/*physicaldevice")  -- returns all sub-items (<physicaldevice> under <c4soap><physicaldevices></></>)
        -- For each <physicaldevice, the attributes of the uuid/address/etc. are in (k,v) table, v.xarg.uuid/v.xarg.address/etc.
    -- XMLHelper.XMLPath(t, "#attr")  -- returns attribute named 'attr' of <c4soap attr="Value1">InnerText</c4soap>
    -- NOTE: Requires that a *table* be passed in, not the actual XML...  XMLHelper.XMLToTable is expected...

    if (strWalk == nil) then return nil end

    local pos = string.find(strWalk, "([/])")
    if (pos ~= nil) then 
      look = string.sub(strWalk, 1, pos - 1)
      walk = string.sub(strWalk, pos + 1)
    else
      look = strWalk
      walk = nil
    end

    local stype = string.sub(look, 1, 1)
    if ((stype == '^') or (stype == '#') or (stype == '*')) then
      look = string.sub(look, 2)
    else
      stype = ""
    end

    if (stype == '#') then
      if (type(tXML.xarg[look]) == "string") then
        return tXML.xarg[look]
      else
        return nil
      end
    end

    local i = 1
    while (tXML[i] ~= nil) do
      if (tXML[i].label == look) then
        if (stype == '*') then
          table.insert(out, tXML[i])
        elseif (stype == '^') then
          if (type(tXML[i][1]) == "string") then
            return tXML[i][1]
          else
            return nil
          end
        else
          -- stype == "" -- next substring...
          return(XMLHelper.XMLPath(tXML[i], walk))
        end
      end
      i = i + 1
    end

    if (#out > 0) then
      return out
    else
      return nil
    end
  end

end
----------------------------------------------------------
-----------------  END XML Helper Module   ------------------
----------------------------------------------------------
