--[[EXEC_LUA]]

--[[
KNX Heat or Cool Thermostat Driver. Allows homeowner to control 'seasonal switchover' system. Used with KNX Master driver.
Copyright 2021 Wirepath Home Systems, LLC. All rights reserved.
]]

require 'XMLHelper'
require 'Languages'

KNX_BINDING = 1
PROXY_BINDING = 5001

PROP_VISIBLE = 0
PROP_HIDDEN = 1

F_RESOLUTIONS = {}
F_RESOLUTIONS["1"] = "1"
F_RESOLUTIONS[".5"] = ".5"
F_RESOLUTIONS[".1"] = ".2"

PRX_CMD = {}
EX_CMD = {}

gModeList = {
  Heating = {
    Allowed = "", -- Mode list string. Cached here to preserve the order of modes in Mode List property
    Modes = {}, -- Name, Id, Type, DisplayName, Address
    Ids = {},
  },
  Cooling = {
    Allowed = "",-- Mode list string. Cached here to preserve the order of modes in Mode List property
    Modes = {}, -- Name, Id, Type, DisplayName, Address
    Ids = {},
  },
}

MODE_OFF_ID = -1
MODE_OFF_NAME = "Off"

VENTILATION_PROPERTIES = {
  "Ventilation Set Address",
  "Ventilation Status Address",
  "Ventilation Mode List",
  "Ventilation Mode DPT",
  "Ventilation Auto/Manual Set Address",
  "Ventilation Auto/Manual Status Address",
  "Ventilation Auto/Manual Operation",
}

function GetModes(season)
  return gModeList[season]
end

function GetCurrentModeList()
  return GetModes(gSeason)
end

function GetCurrentModeId(modeName)
  return GetCurrentModeList().Modes[modeName].Id
end

function GetCurrentModeIds()
  return GetCurrentModeList().Ids
end

function GetModeById(season, modeId)
  return GetModes(season).Ids[tonumber(modeId)]
end

-- Test function, used to issue a read request to KNX bus on specified Group Address, to test incoming KNX data...
-- Example: RequestStatus(Properties["Temperature Status Address"])
--   Causes Temperature to be requested and then sent to this driver...
function RequestStatus(ga)
  MySendToProxy(KNX_BINDING, "REQUEST_STATUS", {GA = ga})
end

function RequestStatusAll()
  for k,v in pairs(Properties) do if (k:find("Status") and (trim(v) ~= "")) then RequestStatus(v) end end
end

function GetSeasonPropertyName(strProperty, season)
  local s = season or gSeason

  return s .. " " .. strProperty
end

function GetSeasonPropertyValue(strProperty, season)
  local prop = GetSeasonPropertyName(strProperty, season)

  return trim(Properties[prop])
end

function GetPropertyTrimmed(prop)
  local val = Properties[prop]

  return val and trim(val)
end

function HasControlOnOff(season)
  return GetSeasonPropertyValue("Control On/Off Address", season) ~= ""
end

function ControlOnOffInverted(season)
  return GetSeasonPropertyValue("On/Off Values Inverted", season) == "Yes"
end

function HasFanAutoManual()
  return GetSeasonPropertyValue('Ventilation Auto/Manual Set Address') ~= ''
end

function IsFanOperationNormal()
  return GetSeasonPropertyValue('Ventilation Auto/Manual Operation') == '1 = Automatic, 0 = Manual'
end

function GetFanAutoValue()
  return IsFanOperationNormal() and 1 or 0
end

function GetFanManualValue()
return IsFanOperationNormal() and 0 or 1
end

function OnBehaviorWhenOffChanged(season, prop)
  local PROP_VISIBILITY = { [true] = PROP_VISIBLE, [false] = PROP_HIDDEN}
  local showMode = prop:find("Mode") ~= nil
  local showSetpoint = prop:find("Setpoint") ~= nil
  local showOnOff = prop == "Turn Off"

  C4:SetPropertyAttribs(GetSeasonPropertyName("Mode When Not In Use", season), PROP_VISIBILITY[showMode])
  C4:SetPropertyAttribs(GetSeasonPropertyName("Setpoint When Not In Use", season), PROP_VISIBILITY[showSetpoint])
end

function SetPropertyVisibility(propertyName, bVisible)
  if (Properties[propertyName] ~= nil) then
    C4:SetPropertyAttribs(propertyName, bVisible and 0 or 1)
  end
end

function OnPropertyChanged(strProperty)
  local prop = Properties[strProperty]
  if (strProperty == "Debug Mode") then
    if ((gDbgTimer or 0) > 0) then gDbgTimer = C4:KillTimer(gDbgTimer) end
    gDbgPrint, gDbgLog = (prop:find("Print") ~= nil), (prop:find("Log") ~= nil)
    if (prop == "Off") then return end
    gDbgTimer = C4:AddTimer(8*60, "MINUTES")
    dbg("Started Debug timer for 8 hours")
  elseif (strProperty == GetSeasonPropertyName("Ventilation Mode List") or
    strProperty == GetSeasonPropertyName("Ventilation Auto/Manual Set Address")) then
      gFanModeList, _gFanModeList = ParseFanModeList()
  elseif (strProperty:find("Mode List")) then
    local season = strProperty:match("(.+) Mode List")
    OnModeListPropertyChanged(season)
  elseif (strProperty:find("Mode 1%-Bit Individual Address")) then
    local season = strProperty:match("(.+) Mode 1%-Bit Individual Address")
    OnModeListPropertyChanged(season)
  elseif (strProperty == "Preset Scheduling") then
    SetupPresetScheduling()
  elseif (strProperty:find("Ventilation Control")) then
    local season = strProperty:match("(.+) Ventilation Control")
    local fanDisabled = prop == "Off"

    for k,v in pairs(VENTILATION_PROPERTIES) do
      C4:SetPropertyAttribs(GetSeasonPropertyName(v, season), fanDisabled and PROP_HIDDEN or PROP_VISIBLE)
    end

    if fanDisabled then
      if (season == gSeason) then
        MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = " "})
      end
    else
      if (season == gSeason) then
        OnPropertyChanged(GetSeasonPropertyName("Ventilation Mode List"))
      end
    end
  elseif (strProperty:find("Mode Temperature Ranges")) then
    local season = strProperty:match("(.+) Mode Temperature Ranges")
    ParseModeRanges(season)
  elseif (strProperty == "Hold Modes") then
    if (prop == "Off") then
      MySendToProxy(PROXY_BINDING, "ALLOWED_HOLD_MODES_CHANGED", {MODES = " "})
    else
      MySendToProxy(PROXY_BINDING, "ALLOWED_HOLD_MODES_CHANGED", {MODES = "Off,2 Hours,Until Next,Permanent"})
    end
  elseif (strProperty == "Setpoint Resolution") then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_RESOLUTION_C = prop})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_RESOLUTION_F = F_RESOLUTIONS[prop]})
  elseif (strProperty == "Temperature Resolution") then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {TEMPERATURE_RESOLUTION_C = prop})
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {TEMPERATURE_RESOLUTION_F = F_RESOLUTIONS[prop]})
  elseif (strProperty:find("Behavior When Not In Use")) then
    local season = strProperty:match("(.*) Behavior When Not In Use")

    OnBehaviorWhenOffChanged(season, prop)

  elseif strProperty == "Heating Mode Status DPT" then
    OnHVACModeStatusDPTChanged("Heating")
  elseif strProperty == "Cooling Mode Status DPT" then
    OnHVACModeStatusDPTChanged("Cooling")
  elseif (strProperty:find("Control On/Off Address")) then
    local season = strProperty:match("(.+) Control On/Off Address")
    OnModeListPropertyChanged(season)
  elseif (strProperty:find("Mode DPT")) then
    local season = strProperty:match("(.+) Mode DPT")
    local bitwise = prop == '1-Bit addresses for each mode'
    SetPropertyVisibility(GetSeasonPropertyName('Mode Set Address', season), not bitwise)
    SetPropertyVisibility(GetSeasonPropertyName('Mode 1-Bit Individual Address', season), bitwise)

    OnModeListPropertyChanged(season)
  end

  AddGroupItemsDelayed(5)

  -- Group Address Cleared, clear cached KNX data
  if trim(prop) == "" then ClearKNXData(strProperty) end
end

function OnHVACModeStatusDPTChanged(season)
  local dpt = GetSeasonPropertyValue("Mode Status DPT", season)
  local hvacStatus = dpt == "HVAC Status (bitwise)"

  SetPropertyVisibility(GetSeasonPropertyName("HVAC Status Address", season), hvacStatus)
  SetPropertyVisibility(GetSeasonPropertyName("Mode Status Address", season), not hvacStatus)
end

function OnModeListPropertyChanged(season)
  local newModeList = ParseModeList(season)
  local oldModeList = GetModes(season)

  -- Rewrite language files
  local poRewrites = {}
  for _, mode in pairs(newModeList.Modes) do
    table.insert(poRewrites, {msgid = mode.DisplayName, msgstr = mode.Name})
  end
  dbg("Updating language files in knx_heat_or_cool_thermostat/www/languages/")
  RewritePOFiles(poRewrites, "knx_heat_or_cool_thermostat/www/languages/")

  C4:UpdatePropertyList(GetSeasonPropertyName("Mode When Not In Use", season), newModeList.Allowed)

  local modeOff = GetSeasonPropertyValue("Mode When Not In Use", season)
  -- Updating "Mode When Off" list will clear selected mode. Check if selected mode is still present
  -- and update "Mode When Off" list
  if (newModeList.Modes[modeOff]) then
    C4:UpdateProperty(GetSeasonPropertyName("Mode When Not In Use", season), modeOff)
  else
    local oldModeData = oldModeList.Modes[modeOff]
    if oldModeData then
      local newMode = newModeList.Ids[oldModeData.Id]
      if newMode then
        C4:UpdateProperty(GetSeasonPropertyName("Mode When Not In Use", season), newMode.DisplayName)
      end
    end
  end

  gModeList[season] = newModeList
  -- Also update temperature ranges
  ParseModeRanges(season)
  if (season == gSeason) then
    UpdateModeList()
    -- Setup Preset fields
    SetupPresetScheduling()
  end

  -- If mode names changed, update presets to use new names
  UpdatePresetValues(season, newModeList, oldModeList)
end

function UpdatePresetValues(season, newModeList, oldModeList)
  -- If mode names changed, update presets to use new names
  if (not gPresets) or (not gPresets[season]) then return end

  for name, data in pairs(gPresets[season]) do
    if data.hvac_mode and not newModeList.Modes[data.hvac_mode] then
      local oldModeData = oldModeList.Modes[data.hvac_mode]
      if oldModeData then
        local newMode = newModeList.Ids[oldModeData.Id]
        if newMode then
          if (season == gSeason) then
            -- Copy other preset fields, but use new hvac mode name
            local presetFields = {}
            for id, value in pairs(data) do
              if (id == "hvac_mode") then
                presetFields[id] = newMode.DisplayName
              else
                presetFields[id] = value
              end
            end
            PRESET_MODIFY(name, nil, presetFields)
          else
            data.hvac_mode = newMode.DisplayName
          end
        end
      end
    end
  end
end

function PRESET_MODIFY(strName, strNewName, tPresetFields)
  local tParams = { NAME = strName, NEW_NAME = strNewName }
  local presetFields = ""
  for name, value in pairs(tPresetFields) do
    presetFields = presetFields .. '<field id="' .. name .. '" value="' .. value .. '" />'
  end

  tParams.PRESET_FIELDS = '<preset_fields>' .. presetFields .. '</preset_fields>'
  MySendToTStatDevice("PRESET_MODIFY", tParams)
end

function OnDriverDestroyed()
  gDbgTimer = C4:KillTimer(gDbgTimer or 0)
end

function dbg(strDebugText)
  if (gDbgPrint) then print (strDebugText) end
  if (gDbgLog) then C4:ErrorLog("KNX Thermostat: " .. strDebugText) end
end


function formatParams(tParams)
  tParams = tParams or {}
  local out = {}
  for k,v in pairs(tParams) do
    if (type(v) == "string") then
      table.insert(out, k .. ': "' .. tostring(v):sub(1, 1000) .. '"')
    else
      table.insert(out, k .. ": " .. tostring(v):sub(1, 1000))
    end
  end
  return table.concat(out, ", ")
end


function OnTimerExpired(idTimer)
  if (idTimer == gDbgTimer) then
    dbg("Turning Debug Mode Off (timer expired)")
    C4:UpdateProperty("Debug Mode", "Off")
    OnPropertyChanged("Debug Mode")
    gDbgTimer = C4:KillTimer(gDbgTimer)
  end

  -- Kill any 'stray' timers (should not be needed)
  C4:KillTimer(idTimer)
end


function trim(s)
  s = s or ""
  return (s:gsub("^%s*(.-)%s*$", "%1")) -- from PiL2 20.4
end

function split(str, delimiter)
  local ret = {}
  for m in string.gmatch(str, '[^' .. delimiter .. ']+') do table.insert(ret, trim(m)) end
  return ret
end

-- Customized version of tonumber... Should work regardless of locale...
function tonumber_loc(str, base)
  if (type(str) ~= "string") then str = tostring(str) end

  local s = str:gsub(",", ".") -- Assume US Locale decimal separator
  local num = tonumber(s, base)
  if (num == nil) then
    s = str:gsub("%.", ",") -- Non-US Locale decimal separator
    num = tonumber(s, base)
  end
  return num
end

function table.empty(t)
  return next(t) == nil
end

function SendToKNX(tParams)
  tParams = tParams or {}
  -- Use 'canonical' Network Driver name for DPT and GA...
  tParams.DATA_POINT_TYPE = tParams.DATA_POINT_TYPE or tParams.DPT
  tParams.DPT = nil
  tParams.GROUP_ADDRESS = tParams.GROUP_ADDRESS or tParams.GA
  tParams.GA = nil

  MySendToProxy(KNX_BINDING, "SEND_TO_KNX", tParams)
end

function SendToKNXProp(prop, dpt, tParams)
  local ga = GetPropertyTrimmed(prop)

  if (ga ~= "") then
    tParams = tParams or {}
    tParams.DATA_POINT_TYPE = dpt
    tParams.GROUP_ADDRESS = ga
    SendToKNX(tParams)
  else
    dbg("Not sending data to KNX - Group Address is not specified for property " .. prop)
  end
end

function SendKNXPropValue(propertyName, dpt, value)
  SendToKNXProp(propertyName, dpt, {VALUE = value})
  SetKNXData(propertyName, value)
end

----------------------- THERMOSTAT HELPER FUNCTIONS -------------------------
FEEDBACK_ITEMS = {}
FEEDBACK_ITEMS["Temperature Status Address"] = {
  DPT = "DPT_9",
}

FEEDBACK_ITEMS["Setpoint Status Address"] = {
  DPT = "DPT_9",
}

FEEDBACK_ITEMS["Mode Status Address"] = {
  GET_DPT = function(season)
    return select(3, (GetSeasonPropertyValue("Mode Status DPT") or "DPT_1 "):find("(DPT_%d) .*"))
  end,
  GET_ENABLED = function(season)
    return GetSeasonPropertyValue("Mode Status DPT") ~= "HVAC Status (bitwise)"
  end
}

FEEDBACK_ITEMS["HVAC Status Address"] = {
  DPT = "DPT_5",
  GET_ENABLED = function(season)
    return GetSeasonPropertyValue("Mode Status DPT") == "HVAC Status (bitwise)"
  end
}

FEEDBACK_ITEMS["Status Address"] = {
  GET_DPT = function(season)
    return select(3, (GetSeasonPropertyValue("Status DPT") or "DPT_1 "):find("(DPT_%d) .*"))
  end,
}

FEEDBACK_ITEMS["Ventilation Status Address"] = {
  GET_DPT = function(season)
    return select(3, (GetSeasonPropertyValue("Ventilation Mode DPT") or "DPT_5 "):find("(DPT_%d) .*"))
  end,
}

FEEDBACK_ITEMS["Ventilation Auto/Manual Status Address"] = {
  DPT = "DPT_1",
}

FEEDBACK_ITEMS["Control On/Off Address"] = {
  DPT = "DPT_1",
}

-- Status group addresses common for both heating and cooling system
FEEDBACK_ITEMS_COMMON = {}
FEEDBACK_ITEMS_COMMON["Heating/Cooling Changeover Address"] = {
  DPT = "DPT_1",
}


function GetKNXHVACModeRaw()
  return GetKNXData(GetSeasonPropertyName("Mode Status Address"))
end

HVAC_MODE_STATUS_HANDLERS = {
  ["DPT_1 (1-Bit)"] = GetKNXHVACModeRaw,

  ["DPT_5 (1-Byte)"] = GetKNXHVACModeRaw,

  ["HVAC Status (bitwise)"] = function()
    local value = tonumber(GetKNXData(GetSeasonPropertyName("HVAC Status Address")))
    if not value then return nil end

    for i = 0, 3 do
      if bit.band(value, bit.lshift(0x1, i)) ~= 0x0 then
        -- For now, assume following bit values:
        -- [0] = Comfort, [1] = Standby, [2] = Economy, [3] = Protection
        -- This assumption for now seems reasonable, since most thermostat use this bit organization
        -- In case of a thermostat with different bit organization, add a property
          return i + 1
      end
    end

    return nil
  end,
}

-- Data from KNX. Key - Group Address property name
KNX_DATA = {}

function GetKNXData(propertyName)
  return KNX_DATA[propertyName]
end

function SetKNXData(propertyName, value)
  KNX_DATA[propertyName] = value
end

function ClearKNXData(propertyName)
  SetKNXData(propertyName, nil)
end

function GetCurrentHVACMode()
  local modeId = HVAC_MODE_STATUS_HANDLERS[GetSeasonPropertyValue("Mode Status DPT")]()
  if not modeId then
    dbg("Unable to get current HVAC mode - no data from KNX.")
    return nil
  end

  return GetCurrentModeIds()[tonumber(modeId)]
end

function IsThermostatOn()
  if HasControlOnOff() then
    local value = tonumber(GetKNXData(GetSeasonPropertyName("Control On/Off Address")))
    if not value then
      dbg("Unable to determine Control On/Off value - no data from KNX.")
      return true
    end

    if ControlOnOffInverted() then value = 1 - value end
    return value == 1
  else
    return true
  end
end


function OnBindingChanged(idBinding, strClass, bConnected)
  if (strClass == "KNX_DEVICE") and (bConnected) then
    dbg("KNX Binding Changed. Sending Group Addresses...")
    AddGroupItemsDelayed(10)
  end
end

function AddGroupItem(propName, data)
  local prop = trim(Properties[propName])
  local enabled
  if (data.GET_ENABLED ~= nil) and (type(data.GET_ENABLED) == "function") then
    enabled = data.GET_ENABLED(gSeason)
  else
    enabled = true
  end

  if (prop == "") or not enabled then return end

  local dpt
  if (data.GET_DPT and type(data.GET_DPT) == "function") then
    dpt = data.GET_DPT(gSeason)
  else
    dpt = data.DPT
  end

  MySendToProxy(KNX_BINDING, "ADD_GROUP_ITEM", {GROUP_ADDRESS = prop, DEVICE_ID = C4:GetDeviceID(), PROPERTY = propName, DATA_POINT_TYPE = dpt})
end


function AddGroupItems()
  RemoveGroupItems()

  for item, data in pairs(FEEDBACK_ITEMS) do AddGroupItem(GetSeasonPropertyName(item), data) end
  for item, data in pairs(FEEDBACK_ITEMS_COMMON) do AddGroupItem(item, data) end
end

function AddGroupItemsDelayed(delaySeconds)
  OneShotTimer.Add(delaySeconds, "SECONDS", AddGroupItems, "AGI")
end


function RemoveGroupItems()
  MySendToProxy(KNX_BINDING, "CLEAR_GROUP_ITEMS", {DEVICE_ID = C4:GetDeviceID()})
end


function FtoC(valueF)
  -- Verify sane input...
  valueF = tonumber_loc(valueF) or 0

  -- °C = (°F - 32) * 5/9
  local valueC = ((valueF - 32) * 5 / 9)

  -- No need to round...
  return valueC
end

function CtoF(valueC)
  valueC = tonumber_loc(valueC) or 0

  local tempF = (valueC * 1.8) + 32

  return math.floor(tempF + 0.5)
end

FROM_BINDING = {}
FROM_BINDING[KNX_BINDING] = "<---- KNX "
FROM_BINDING[PROXY_BINDING] = "<---- Proxy"

function GetHVACModeListId(season)
  local ret = {}
  local ml = GetSeasonPropertyValue("Mode List", season) .. ","

  for item in ml:gfind("(.-),") do
    local _, _, name, id = item:find("(.+)=(.+)")
    name, id = trim(name), tonumber(trim(id))
    table.insert(ret, {Name = name, Id = id})
  end

  return ret
end

function SetHVACModeId(season, modeData, dpt)
  SendKNXPropValue(GetSeasonPropertyName("Mode Set Address", season), dpt, modeData.Id)
end

HVAC_MODE_HANDLERS = {}

HVAC_MODE_HANDLERS['DPT_1 (1-Bit)'] = {
  GetHVACModeList = GetHVACModeListId,

  SET = function(season, modeData)
    SetHVACModeId(season, modeData, 'DPT_1')
  end,
}

HVAC_MODE_HANDLERS['DPT_5 (1-Byte)'] = {
  GetHVACModeList = GetHVACModeListId,

  SET = function(season, modeData)
    SetHVACModeId(season, modeData, 'DPT_5')
  end,
}

HVAC_MODE_HANDLERS['1-Bit addresses for each mode'] = {
  GetHVACModeList = function(season)
    local bitwiseModes = {}
    local ml = GetSeasonPropertyValue("Mode 1-Bit Individual Address", season)
    for _, m in ipairs(split(ml, ',')) do
      local name, ga = m:match("(.-)=(%d+/%d+/%d+)")
      ga = ga and trim(ga)
      if name and ga then table.insert(bitwiseModes, {Name = name, Address = ga}) end
    end

    -- Merge with HVAC mode list in status modes
    local modes = GetHVACModeListId(season)
    for _, mode in ipairs(modes) do
      -- Find bitwise mode
      local baseName = GetHVACModeNameBase(mode.Name)
      for _, m in ipairs(bitwiseModes) do
        if GetHVACModeNameBase(m.Name) == baseName then
          mode.Address = m.Address
          break
        end
      end
    end

    return modes
  end,

  SET = function(season, modeData)
    if modeData.Address then
      SendToKNX({GROUP_ADDRESS = modeData.Address, DATA_POINT_TYPE = 'DPT_1', VALUE = 1})
    end

    for mode, data in pairs(GetCurrentModeList().Modes) do
      if data.Address and data.Address ~= modeData.Address then
        SendToKNX({GROUP_ADDRESS = data.Address, DATA_POINT_TYPE = 'DPT_1', VALUE = 0})
      end
    end
  end,
}

function ParseModeList(season)
  local ModeList = {Modes = {}, Ids = {}, Allowed = ""}
  local tModes = HVAC_MODE_HANDLERS[GetSeasonPropertyValue('Mode DPT', season)].GetHVACModeList(season)
  local allowedModes = {}

  for _, mode in ipairs(tModes) do
    local modename = mode.Name
    local orig = modename

    -- Automatically insert _Heat if no 'mode-type' in mode name...
    local orig, modeType = GetHVACModeNameBase(modename)
    if (modeType) then
      modename = orig .. "_" .. modeType
    elseif not (modename:find('Heat') or modename:find('Cool') or modename:find('Auto')
        or modename:find('Off')) then
      modeType = (season == "Heating") and "Heat" or "Cool"
      modename = modename .. "_" .. modeType
    end

    -- local modeData = {Name = orig, Id = id, Type = modeType, DisplayName = modename}
    mode.Type = modeType
    mode.Name = orig
    mode.DisplayName = modename
    ModeList.Modes[modename] = mode
    if mode.Id then
      ModeList.Ids[mode.Id] = mode
    end
    table.insert(allowedModes, modename)
  end

  -- If On/Off address is used, add the Off mode to the list
  if HasControlOnOff(season) then
    local modeData = {Name = MODE_OFF_NAME, Id = MODE_OFF_ID, Type = "Off", DisplayName = MODE_OFF_NAME}
    ModeList.Modes[MODE_OFF_NAME] = modeData
    ModeList.Ids[MODE_OFF_ID] = modeData
    table.insert(allowedModes, MODE_OFF_NAME)
  end

  ModeList.Allowed = table.concat(allowedModes, ",")
  return ModeList
end

function GetHVACModeNameBase(modename)
  local first, last, modeType = modename:find('%[(.+)%]')
  if (first) then
    return modename:sub(0,first-1) .. modename:sub(last+1), modeType
  else
    return modename
  end
end

function UpdateModeList()
  MySendToProxy(PROXY_BINDING, "ALLOWED_HVAC_MODES_CHANGED", {MODES = GetCurrentModeList().Allowed})
end


function ParseFanModeList()
  local fanControl = GetSeasonPropertyValue("Ventilation Control")
  local ModeList = {}
  local _ModeList = {}
  local strFanModes

  if (fanControl ~= "Off") then
    local ml = GetSeasonPropertyValue("Ventilation Mode List") .. ","
    local allowed = {}

    for item in ml:gfind("(.-),") do
      local _,_,modename,id = item:find("(.+)=(.+)")
      modename,id = trim(modename), trim(id)
      ModeList[modename] = id
      _ModeList[id] = modename
      table.insert(allowed, modename)
    end

    -- Check if Automatic/Manual Address is specified
    if HasFanAutoManual() then table.insert(allowed, 1, 'Auto') end
    strFanModes = table.concat(allowed, ",")
  else
    strFanModes = " "
  end

  MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = strFanModes})
  return ModeList, _ModeList
end

function UpdateCurrentFanMode()
  local auto
  if HasFanAutoManual() then
    local autoValue = GetKNXData(GetSeasonPropertyName('Ventilation Auto/Manual Status Address')) or
      GetKNXData(GetSeasonPropertyName('Ventilation Auto/Manual Set Address'))
    if autoValue and tonumber(autoValue) == GetFanAutoValue() then auto = true end
  end

  local mode
  if auto then
    mode = 'Auto'
  else
    mode = _gFanModeList[GetKNXData(GetSeasonPropertyName('Ventilation Status Address')) or
      GetKNXData(GetSeasonPropertyName('Ventilation Set Address'))]
  end

  MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", { MODE = mode })
end


function ParseModeRanges(season)
  local ml = GetSeasonPropertyValue("Mode Temperature Ranges", season) .. ","

  for item in ml:gfind("(.-),") do
    local _,_,num,range = item:find("(.+)=(.+)")
    num, range = tonumber(trim(num)), trim(range)
    local _,_,min,max = range:find("(.+)-(.+)")

    local modeData = GetModeById(season, num)

    if modeData and min then modeData.min = trim(min) end
    if modeData and max then modeData.max = trim(max) end
  end
end


function SetRanges(mode)
  local mode = GetCurrentModeList().Modes[mode]
  if (not mode) then return end

  if (mode.min ~= nil ) then
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_MIN_C = mode.min, SINGLE_SETPOINT_MIN_F = tostring(CtoF(mode.min))})
    gCurrentMin = tonumber_loc(mode.min) or -255
  end
  if (mode.max ~= nil and tonumber_loc(mode.max) ~= nil) then
    local max = tonumber_loc(mode.max)
    if max == gCurrentMin then max = max + 1 end
    MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {SINGLE_SETPOINT_MAX_C = tostring(max), SINGLE_SETPOINT_MAX_F = tostring(CtoF(max))})
    gCurrentMax = max or -255
  end
end

function GetThermostatProxyId()
  return next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), PROXY_BINDING))
end

function DeletePresets()
  gDeletingPresets = true
  -- Get Bound Proxy's Device ID
  local proxyid = GetThermostatProxyId()
  local presetxml = C4:SendUIRequest(proxyid, "GET_PRESETS", {})

  -- For each preset, delete it.
  local presets = XMLHelper.XMLToTable(presetxml)
  for i, v in ipairs(presets[1]) do
    dbg("Deleting Preset " .. v.xarg.name )
    MySendToTStatDevice("PRESET_DELETE", {NAME = v.xarg.name})
  end
  gDeletingPresets = false
  ClearPreset()
end


function AddModePresets()
  for k, v in pairs(GetCurrentModeList().Modes) do
    local preset_fields = '<preset_fields><field id="hvac_mode" value="' .. k .. '" /></preset_fields>'
    MySendToTStatDevice("PRESET_ADD", {NAME = v.Name, PRESET_FIELDS = preset_fields})
  end
end

function LoadPresets()
  for name,fields in pairs(gPresets[gSeason]) do
    local preset_fields = '<preset_fields>'
    for id,value in pairs(fields) do
      preset_fields = preset_fields .. '<field id="' .. id .. '" value="' .. value .. '"/>'
    end
    preset_fields = preset_fields .. "</preset_fields>"
    dbg("Load Presets: " .. preset_fields)
    MySendToTStatDevice("PRESET_ADD", { NAME = name, PRESET_FIELDS = preset_fields })
  end
end

function SaveEvents()
  -- Get Bound Proxy's Device ID.
  local proxyId = GetThermostatProxyId()
  local eventXml = C4:SendUIRequest(proxyId, "GET_PRESET_EVENTS", {})

  dbg(eventXml)

  local events = C4:ParseXml(eventXml or "")
  local seasonEvents = {}

  if not events then return end

  for _, event in pairs(events.ChildNodes) do
    local attrs = event.Attributes

    table.insert(seasonEvents, {PRESET = attrs.preset, WEEKDAY = attrs.weekday, HOUR = attrs.hour, MINUTE = attrs.minute })
  end

  for i =1, #seasonEvents do
    formatParams(seasonEvents[i])
  end
  gEvents[gSeason] = seasonEvents
end

function LoadEvents()
  local events = gEvents[gSeason]

  for i = 1, #events do
    MySendToTStatDevice("PRESET_EVENT_ADD", events[i])
  end
end

function GetRangeMinMax()
  -- Find min/max of all min/max values...
  local min, max = 2000, -2000
  for _,v in pairs(GetCurrentModeList().Modes) do

    if (v.min and tonumber_loc(v.min) < min) then min = tonumber_loc(v.min) end
    if (v.max and tonumber_loc(v.max) > max) then max = tonumber_loc(v.max) end
  end
  return min, max
end


function SetPresetFields(preset_fields)
  dbg(preset_fields)
  MySendToProxy(PROXY_BINDING, "PRESET_FIELDS_CHANGED", {XML = preset_fields})
  PersistData.PresetFields = preset_fields
end

function SetupPresetFields()
  local presetScheduling = Properties["Preset Scheduling"]
  local enablePresetScheduling =  presetScheduling ~= "Off"

  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET = enablePresetScheduling})
  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET_SCHEDULE = enablePresetScheduling})

  if (not enablePresetScheduling) then
    SetPresetFields("")
    gPresets[gSeason] = {}
    return
  end

  local min, max = GetRangeMinMax()
  local minF, maxF = CtoF(min), CtoF(max)
  local res = Properties["Setpoint Resolution"] or ".5"

  local preset_fields = '<preset_fields>'

  if (presetScheduling == "Mode Only") then
    preset_fields = preset_fields .. '<field id="hvac_mode" type="list" label="Mode"><list>'
    for k in pairs(GetCurrentModeList().Modes) do
      preset_fields = preset_fields .. '<item text="' .. k .. '" value="' .. k .. '" />'
    end
    preset_fields = preset_fields .. '</list></field>'
  elseif (presetScheduling == "Mode+Temp") then
    preset_fields = preset_fields .. '<field id="hvac_mode" type="list" label="Mode"><list>'
    for k in pairs(GetCurrentModeList().Modes) do
      preset_fields = preset_fields .. '<item text="' .. k .. '" value="' .. k .. '" />'
    end
    preset_fields = preset_fields .. '</list></field>'
    preset_fields = preset_fields .. '<field id="single_setpoint_c" type="number" label="Setpoint" min="' .. min .. '" max="' .. max .. '" res="' .. res .. '" />'
    preset_fields = preset_fields .. '<field id="single_setpoint_f" type="number" label="Setpoint" min="' .. minF .. '" max="' .. maxF .. '" res="' .. res .. '" />'
  elseif (presetScheduling == "Temperature Only") then
    preset_fields = preset_fields .. '<field id="single_setpoint_c" type="number" label="Setpoint" min="' .. min .. '" max="' .. max .. '" res="' .. res .. '" />'
    preset_fields = preset_fields .. '<field id="single_setpoint_f" type="number" label="Setpoint" min="' .. minF .. '" max="' .. maxF .. '" res="' .. res .. '" />'
  end

  preset_fields = preset_fields .. '</preset_fields>'
  SetPresetFields(preset_fields)

  if ((Properties["Preset Scheduling"] == "Mode Only") and table.empty(gPresets[gSeason])) then
    AddModePresets()
  end
end

function SetupPresetScheduling()
  gLastScheduledPreset = ""

  SetupPresetFields()
end

function LoadPresetScheduling()
  local preset_scheduling = Properties["Preset Scheduling"]

  if (preset_scheduling == "Off") then return end

  SetupPresetFields()
  LoadPresets()
end


function RestorePresetScheduling()
  local preset_scheduling = Properties["Preset Scheduling"] ~= "Off"

  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET = preset_scheduling})
  MySendToProxy(PROXY_BINDING, "DYNAMIC_CAPABILITIES_CHANGED", {CAN_PRESET_SCHEDULE = preset_scheduling})

  if (not preset_scheduling ) then
    DeletePresets()
    PersistData.PresetFields = ""
  else
    PersistData = PersistData or {}
    local preset_fields = PersistData.PresetFields or ""
    if (preset_fields ~= "") then
      MySendToProxy(PROXY_BINDING, "PRESET_FIELDS_CHANGED", {XML = preset_fields})
    end
  end
end

function ClearPreset()
  MySendToProxy(PROXY_BINDING, "PRESET_CHANGED", {NAME = ""})
end

function SetupExtras()
  local xml = [[
    <extras_setup><extra>
      <section label="Season Selection">
        <object type="list" id="SeasonSwitch" label="Season" command="EXTRAS_CHANGE_SEASON">
          <list maxselections="1" minselections="1">
            <item text="Heating" value="Heating"/>
            <item text="Cooling" value="Cooling"/>
          </list>
        </object>
      </section>
    </extra></extras_setup>
  ]]

  MySendToProxy(PROXY_BINDING, "EXTRAS_SETUP_CHANGED", { XML = xml })
  UpdateSeasonExtra()
end

function UpdateSeasonExtra()
  local xml = '<extras_state><extra><object id="SeasonSwitch" value="' .. gSeason .. '"/></extra></extras_state>'

  MySendToProxy(PROXY_BINDING, "EXTRAS_STATE_CHANGED", {XML = xml})
end

function SetUntilNextHoldMode()
  PRX_CMD.SET_MODE_HOLD({ MODE = "Until Next" })
  ClearPreset()
end

function ReceivedFromProxy(idBinding, strCommand, tParams)
  strCommand = strCommand or ""
  tParams = tParams or {}

  if (not IGNORE_MESSAGES[strCommand]) then
    dbg("[" .. FROM_BINDING[idBinding] .. "]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  end

  if (PRX_CMD[strCommand] ~= nil) then
    PRX_CMD[strCommand](tParams)
  else
    dbg("Unhandled proxy command: " .. strCommand)
  end
end

function PRX_CMD.SET_SCALE(tParams)
  local scale = tParams["SCALE"]
  if (scale == nil) then return end
  gMyScale = scale
  PersistData.Scale = scale
  MySendToProxy(PROXY_BINDING, "SCALE_CHANGED", {SCALE = gMyScale})
end

function PRX_CMD.SET_MODE_FAN(tParams)
  local mode = tParams["MODE"]
  if not mode then return end

  if HasFanAutoManual() then
    SendKNXPropValue(GetSeasonPropertyName("Ventilation Auto/Manual Set Address"),
      "DPT_1", mode == 'Auto' and GetFanAutoValue() or GetFanManualValue())
  end

  if (not HasFanAutoManual()) or mode ~= "Auto" then
    local _, _, dpt = (GetSeasonPropertyValue("Ventilation Mode DPT") or "DPT_5 "):find("(DPT_%d) .*")
    SendKNXPropValue(GetSeasonPropertyName("Ventilation Set Address"), dpt, gFanModeList[mode])
  end

  -- Change UI if no response from KNX for 5 seconds...
  MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = mode})
end

function PRX_CMD.SET_SETPOINT_SINGLE(tParams)
  local value = tonumber_loc(tParams["CELSIUS"]) or -255
  if (value == -255) then return end

  local isPreset = tParams["PRESET"] or false

  -- If we have min/max for current mode, bracket temp within min/max
  gCurrentMin = gCurrentMin or -255
  gCurrentMax = gCurrentMax or -255
  if (gCurrentMin > -255) and (gCurrentMax > -255) then
    if (value < gCurrentMin) then dbg("Setpoint [" .. value .. "] increased into range --> [" .. gCurrentMin .. "]") value = gCurrentMin end
    if (value > gCurrentMax) then dbg("Setpoint [" .. value .. "] decreased into range --> [" .. gCurrentMax .. "]") value = gCurrentMax end
  end

  SendSetpointToKNX(gSeason, value)

  if (not isPreset) then
    -- Only set Hold Until Next if presets are done with Temperature (not for Off or Mode Only)
    local psmode = Properties["Preset Scheduling"] or ""
    if (psmode == "Mode+Temp") or (psmode == "Temperature Only") then
      SetUntilNextHoldMode()
    end
  end

  -- Change UI if no response from KNX for 5 seconds...
  MySendToProxy(PROXY_BINDING, "SINGLE_SETPOINT_CHANGED", {SCALE = "C", SETPOINT = tostring(value)})
end

function PRX_CMD.SET_MODE_HVAC(tParams)
  local mode = tParams["MODE"]
  if not mode then return end

  local isPreset = tParams["PRESET"] or false

  SendHVACModeToKNX(gSeason, mode)
  SetRanges(mode)

  -- Change UI if no response from KNX for 5 seconds...
  MySendToProxy(PROXY_BINDING, "HVAC_MODE_CHANGED", {MODE = mode})

  if (not isPreset) then
    -- Only set Hold Until Next if presets are done with Mode (not for Off or Temperature Only)
    local psmode = Properties["Preset Scheduling"] or ""
    if (psmode == "Mode+Temp") or (psmode == "Mode Only") then
      SetUntilNextHoldMode()
    end
  end
end

function PRX_CMD.SET_PRESETS(tParams)
  if (gDeletingPresets == true) then return end

  -- Capture presets in global table...
  -- Parse <presets> XML for name, setpoints, preset fields...
  local xml = tParams["XML"] or ""
  if (xml == "") then return end

  local presets = XMLHelper.XMLToTable(xml)
  local parsedPresets = {}

  for i = 1, #presets[1] do
    local mypreset = presets[1][i]
    local myargs = mypreset.xarg or {}
    local name = myargs["name"] or ""
    local pfxml = myargs["preset_fields"] or ""

    -- Must find both name and set of preset fields XML to proceed...
    if ((name ~= "") and (pfxml ~= "")) then
      parsedPresets[name] = {}

      -- Convert quoted XML back to XML...
      pfxml = pfxml:gsub("&gt;", ">"):gsub("&lt;", "<"):gsub("&quot;", '"'):gsub("&amp;", "&")

      dbg(pfxml)

      -- For each field in preset_fields, grab the id, value and store for later...
      local pvals = XMLHelper.XMLToTable(pfxml)
      for i = 1, #pvals[1] do
        local mypval = pvals[1][i]
        local mypargs = mypval.xarg or {}
        local id = mypargs["id"] or ""
        local value = mypargs["value"] or ""
        if (id ~= "") and (value ~= "") then
          parsedPresets[name][id] = value
        end
      end
    end
  end

  if gSeason and gPresets then
    gPresets[gSeason] = parsedPresets
    PersistData.Presets = gPresets
  end
end

function PRX_CMD.SET_EVENT(tParams)
  local preset = tParams["NAME"] or tParams["PRESET"] or ""
  if (preset == "") then return end

  -- Hold Modes only effect Scheduled Event Preset changes, not SET_PRESET changes...
  gLastScheduledPreset = preset -- Save Scheduled Presets to 'restore'...

  if (gHoldMode == "Permanent" or gHoldMode == "2 Hours") then
    dbg("Hold Mode " .. gHoldMode .. " prevents schedule change.")
    return
  end
  if (gHoldMode == "Until Next") then
    dbg("Hold Mode 'Until Next' ended.")
    gHoldMode = "Off"
    MySendToProxy(PROXY_BINDING, "HOLD_MODE_CHANGED", {MODE = gHoldMode})
  end

  PRX_CMD.SET_PRESET(tParams)
end

function PRX_CMD.SET_PRESET(tParams)
  local preset = tParams["NAME"] or tParams["PRESET"] or ""

  if (preset == "") then return end

  MySendToProxy(PROXY_BINDING, "PRESET_CHANGED", {NAME = preset})

  local mypreset = gPresets[gSeason][preset]
  if not mypreset then return end

  if (mypreset.hvac_mode ~= nil) then
    PRX_CMD.SET_MODE_HVAC({MODE = mypreset.hvac_mode, PRESET = true})
  end

  local tempval = mypreset.single_setpoint_c or ""
  if (tempval ~= "") then
    OneShotTimer.Add(2, "SECONDS",
      function()
        PRX_CMD.SET_SETPOINT_SINGLE({CELSIUS = tempval, PRESET = true})
      end
      , "SPMC")
  end
end

function PRX_CMD.SET_EVENTS(tParams)
end

function PRX_CMD.SET_MODE_HOLD(tParams)
  gPrevMode = gHoldMode
  gHoldMode = tParams["MODE"] or "Off"
  if (gHoldMode == gPrevMode) then return end

  MySendToProxy(PROXY_BINDING, "HOLD_MODE_CHANGED", tParams)

  -- When Hold Mode is turned off, restore 'current' preset...
  if (gHoldMode == "Off") then
    if ((gLastScheduledPreset or "") ~= "") then
      dbg("Hold Mode Off.  Restoring Preset [" .. gLastScheduledPreset .. "]")
      PRX_CMD.SET_PRESET({ NAME = gLastScheduledPreset } )
    end
  end

  OneShotTimer.Kill("HOLDMODE")

  -- Set a Hold Mode timer for 2 hours mode only... Everything else, no timer needed
  if (gHoldMode == "2 Hours") then
    OneShotTimer.Add(2, "HOURS",
      function()
        PRX_CMD.SET_MODE_HOLD({ MODE = "Off" })
      end
    , "HOLDMODE")
  end
end

function PRX_CMD.EXTRAS_CHANGE_SEASON(tParams)
  local season = tParams.value

  if ((season ~= "Heating") and (season ~= "Cooling")) then return end

  ChangeSeason(season)
end

function SetHeatCoolStatus()
  gHeatStatus = gHeatStatus or 0
  gCoolStatus = gCoolStatus or 0

  if (gHeatStatus > 0 and gSeason == 'Heating') then
    MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Heat"})
  elseif (gCoolStatus > 0 and gSeason == 'Cooling') then
    MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Cool"})
  else
    MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Off"})
  end

end


FILTER = {}
FILTER["ANA"] = 1
FILTER["V1 "] = 1
FILTER["HOLD"] = 1
FILTER["HEAT"] = 1
FILTER["COOL"] = 1
FILTER["DR"] = 1
FILTER["REVERSING"] = 1
FILTER["SCHEDULE"] = 1
FILTER["DEADBAND"] = 1
FILTER["CALIBRATION"] = 1
FILTER["OUTDOOR"] = 1
FILTER["DISPLAY"] = 1
FILTER["VACATION"] = 1
FILTER["HUMIDITY_STATE"] = 1
FILTER["HUMIDITY_MODE"] = 1
FILTER["HUMIDIFY"] = 1


function ListDeviceVariables()
  local out, tins = {}, table.insert
  local proxyDeviceID = next(C4:GetBoundConsumerDevices(C4:GetDeviceID(), PROXY_BINDING))
  tins(out, "------------------- Proxy Variables -------------------")
  for k,v in pairs(C4:GetDeviceVariables(proxyDeviceID)) do
    local ok = true
    for filter in pairs(FILTER) do
      if (v.name:find(filter)) then ok = false end
    end
    if (v.name == "TEMPERATURE") then ok = false end
    if (ok) then tins(out, string.format("%-6d", k) .. v.name .. ": " .. v.value) end
  end
  tins(out, "------------------- Protocol Variables -------------------")
  for k,v in pairs(C4:GetDeviceVariables(C4:GetDeviceID())) do
    tins(out, string.format("%-6d", k) .. v.name .. ": " .. v.value)
  end
  print(table.concat(out, "\n"))
end

function SendHVACModeToKNX(season, mode)
  local modeData = GetModes(season).Modes[mode]
  if (not modeData) then
    dbg("ERROR: Unknown mode " .. mode)
    return
  end

  if modeData.Id ~= MODE_OFF_ID then
    if HasControlOnOff(season) then TurnSeasonOnOff(season, true) end

    HVAC_MODE_HANDLERS[GetSeasonPropertyValue('Mode DPT', season)].SET(season, modeData)
  else
    TurnSeasonOnOff(season, false)
  end
end

function SendSetpointToKNX(season, value)
  -- TODO: What if this value is in Fahrenheit on Room Control / Thermostat device???
  SendKNXPropValue(GetSeasonPropertyName("Setpoint Set Address", season), "DPT_9", tostring(value))
end

function SendChangeoverToKNX()
  local value
  if gSeason == "Heating" then
    value = 1
  else
    value = 0
  end
  SendKNXPropValue("Heating/Cooling Changeover Address", "DPT_1", value)
end

function TurnSeasonOnOff(season, turnOn)
  local ON_OFF = { [true] = 1, [false] = 0}
  local value = ON_OFF[turnOn]

  if ControlOnOffInverted(season) then value = 1 - value end

  SendKNXPropValue(GetSeasonPropertyName("Control On/Off Address", season), "DPT_1", tostring(value))
end

function EnableSeason(season, enable)
  local behavior = GetSeasonPropertyValue("Behavior When Not In Use", season)

  if ((behavior == "Set Mode + Setpoint") or (behavior == "Set Mode")) then
    if enable then return end

    local mode = GetSeasonPropertyValue("Mode When Not In Use", season)

    if mode ~= "" then
      SendHVACModeToKNX(season, mode)

      if (behavior == "Set Mode + Setpoint") then
        local setpoint = GetSeasonPropertyValue("Setpoint When Not In Use", season)

        if (setpoint ~= "") then
          SendSetpointToKNX(season, setpoint)
        end
      end
    end
  elseif (behavior == "Turn Off") then
    TurnSeasonOnOff(season, enable)
  end
end

function OnSeasonChange(season)
  if (gSeason == season) then return end

  local oldSeason = gSeason

  EnableSeason(oldSeason, false)

  SaveEvents()

  DeletePresets()

  PersistData.Season = season
  gSeason = season

  EnableSeason(gSeason, true)

  C4:SetVariable("SEASON", gSeason)
  UpdateSeasonExtra()

  UpdateModeList()
  LoadPresetScheduling()
  LoadEvents()

  gFanModeList, _gFanModeList = ParseFanModeList()

  gHeatStatus = 0
  gCoolStatus = 0
  SetHeatCoolStatus()

  AddGroupItemsDelayed(3)
end

function ChangeSeason(season)
  if (gSeason == season) then return end

  OnSeasonChange(season)

  SendChangeoverToKNX()
end


function ExecuteCommand(strCommand, tParams)
  tParams = tParams or {}
  strCommand = strCommand or ""

  if (strCommand == "") then return end

  if (not IGNORE_MESSAGES[strCommand]) then
    dbg("[<---- ExecC]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  end

  if (EX_CMD[strCommand] ~= nil) then
    EX_CMD[strCommand](tParams)
  else
    dbg("Unhandled command: " .. strCommand)
  end
end

function EX_CMD.OFFLINE(tParams)
  -- Do not report that the device is not connected. It should be possible to create presets
  -- while KNX is offline. Presets are not pushed on the device, so connection is not required
  SetConnected(false)
end

function EX_CMD.ONLINE(tParams)
  SetConnected(true)
end

function EX_CMD.DATA_FROM_KNX(tParams)
  local ga = trim(tParams["GROUP_ADDRESS"])
  local value = tParams["VALUE"] or ""
  if (ga == "") or (value == "") then return end

  -- Find out which Property corresponds to this GA...
  for prop, propval in pairs(Properties) do
    propval = trim(propval)

    if (ga == propval) then
      -- Populate the data from KNX table
      SetKNXData(prop, value)

      if (prop == GetSeasonPropertyName("Setpoint Status Address")) then
        MySendToProxy(PROXY_BINDING, "SINGLE_SETPOINT_CHANGED", {SCALE = "C", SETPOINT = tostring(value)})
      elseif (prop == GetSeasonPropertyName("Temperature Status Address")) then
        SetConnected(true)
        MySendToProxy(PROXY_BINDING, "TEMPERATURE_CHANGED", {SCALE = "C", TEMPERATURE = value})
      elseif (prop == GetSeasonPropertyName("Mode Status Address")) then
        UpdateCurrentHVACMode()
      elseif (prop == GetSeasonPropertyName("Ventilation Status Address")) then
        UpdateCurrentFanMode()
      elseif (prop == "Heating Status Address" and gSeason == 'Heating') then
        gHeatStatus = tonumber_loc(value) or 0
        SetHeatCoolStatus()
      elseif (prop == "Cooling Status Address" and gSeason == 'Cooling') then
        gCoolStatus = tonumber_loc(value) or 0
        SetHeatCoolStatus()
      elseif (prop == "Heating/Cooling Changeover Address") then
        OnSeasonChange((tonumber_loc(value) == 1) and "Heating" or "Cooling")
      elseif (prop == GetSeasonPropertyName("HVAC Status Address")) then
        UpdateCurrentHVACMode()
      elseif (prop == GetSeasonPropertyName("Control On/Off Address")) then
        UpdateCurrentHVACMode()
      elseif (prop == GetSeasonPropertyName("Ventilation Auto/Manual Status Address")) then
        UpdateCurrentFanMode()
      end
    end
  end
end

function UpdateCurrentHVACMode()
  local mode
  local onOff = IsThermostatOn()

  if not onOff then
    mode = GetModeById(gSeason, MODE_OFF_ID)
  else
    mode = GetCurrentHVACMode()
  end

  if mode then
    OneShotTimer.Add(50, "MILLISECONDS", function()
      MySendToProxy(PROXY_BINDING, "HVAC_MODE_CHANGED", {MODE = mode.DisplayName})
      SetRanges(mode.DisplayName)
    end , "HMC")
  end
end

function EX_CMD.FAN_MODE_CHANGED(tParams)
  local fanmode = tParams["MODE"] or ""
  if (fanmode == "") then return end

  fanmode = trim(string.upper(fanmode))
  if (fanmode == "ON") then
    MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = "Off,Fan Timer"})
    MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = "Fan Timer"})
    MySendToProxy(PROXY_BINDING, "FAN_STATE_CHANGED", {STATE = "On"})
  else
    MySendToProxy(PROXY_BINDING, "ALLOWED_FAN_MODES_CHANGED", {MODES = "Off,Fan Timer"})
    MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = "Off"})
    MySendToProxy(PROXY_BINDING, "FAN_STATE_CHANGED", {STATE = "Off"})
  end
end

function EX_CMD.LUA_ACTION(tParams)
    local action = tParams.ACTION or ""

    if (action == "ListVariables") then
      ListDeviceVariables()
    elseif (action == "SwitchSeason") then
      ChangeSeason(tParams.Season)
    end
end

EX_CMD["Switch Season"] = function(tParams)
  ChangeSeason(tParams.Season)
end

-- TODO: Property to turn on/off messages???? -- Also, message list array, command to set/clear messages...
function MySetMessage(strMsg)
  MySendToProxy(PROXY_BINDING, "MESSAGE_CHANGED", {MESSAGE = strMsg})
end


TO_BINDING = {}
TO_BINDING[KNX_BINDING] = "----> KNX  "
TO_BINDING[PROXY_BINDING] = "----> Proxy"

IGNORE_MESSAGES = {}
IGNORE_MESSAGES["UPDATE_SCHEDULE_ENTRIES"] = true


function MySendToProxy(idBinding, strCommand, tParams)

  if (not IGNORE_MESSAGES[strCommand]) then
    dbg("[" .. TO_BINDING[idBinding] .. "]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  end

  C4:SendToProxy(idBinding, strCommand, tParams)
end

function MySendToTStatDevice(strCommand, tParams)
  local proxyid = GetThermostatProxyId()
  dbg("[----> Thermostat Device]: " .. strCommand .. " {" .. formatParams(tParams) .. "}")
  C4:SendToDevice(proxyid, strCommand, tParams)
end


function RealInit()
  gFanModeList, _gFanModeList = ParseFanModeList()

  OnPropertyChanged("Debug Mode")
  OnPropertyChanged("Setpoint Resolution")
  OnPropertyChanged("Temperature Resolution")
  OnPropertyChanged("Hold Modes")

  for _, season in pairs({"Heating", "Cooling"}) do
    for _, prop in ipairs({"Ventilation Control", "Behavior When Not In Use", "Mode List",
      "Mode Temperature Ranges", "Mode Status DPT", "Control On/Off Address", 'Mode DPT'}) do
      OnPropertyChanged(GetSeasonPropertyName(prop, season))
    end
  end

  -- Restore Preset fields to last setting, so we don't lose preset on restart.
  RestorePresetScheduling()

  SetupExtras()

  SendChangeoverToKNX()
end


function OnDriverLateInit()
  PersistData = PersistData or {}
  PersistData.Scale = PersistData.Scale or "C"
  PersistData.Season = PersistData.Season or "Heating"
  PersistData.Presets = PersistData.Presets or { Heating = {}, Cooling = {} }
  PersistData.Events = PersistData.Events or { Heating = {}, Cooling = {} }
  if (PersistData.Scale == "C") or (PersistData.Scale == "CELSIUS") then
    PersistData.Scale = "CELSIUS"
  else
    PersistData.Scale = "FAHRENHEIT"
  end

  gSeason = PersistData.Season
  gPresets = PersistData.Presets or {}
  gEvents = PersistData.Events

  MySendToProxy(PROXY_BINDING, "SCALE_CHANGED", {SCALE = PersistData.Scale})
  SetConnected(true)
  MySendToProxy(PROXY_BINDING, "CONNECTION", {CONNECTED = true})

  OneShotTimer.Add(5, "SECONDS", RealInit, "REAL_LATE_INIT")

  -- Clear any messages...
  MySetMessage("")

  MySendToProxy(PROXY_BINDING, "ALLOWED_HOLD_MODES_CHANGED", {MODES = " "})
  MySendToProxy(PROXY_BINDING, "FAN_MODE_CHANGED", {MODE = "Off"})

  MySendToProxy(PROXY_BINDING, "HVAC_STATE_CHANGED", {STATE = "Off"})
  MySendToProxy(PROXY_BINDING, "FAN_STATE_CHANGED", {STATE = "Off"})

  C4:AddVariable("SEASON", PersistData.Season, "STRING", true)

  -- On versions < 27, Mode DPT property specified both DPT for Mode Set and Mode Status
  -- On version >= 27, Mode Status DPT and HVAC Status Address properties are introduced
  -- Set the Mode Status DPT property based on the version of the driver and Mode DPT properties
  local version = C4:GetDriverConfigInfo("version")
  local prevVersion = PersistData.Version
  if (version ~= prevVersion) then
    PersistData.Version = version
    if (not prevVersion) then
      for _, season in pairs({"Heating", "Cooling"}) do
        local modeDpt = GetSeasonPropertyValue("Mode DPT", season)
        C4:UpdateProperty(GetSeasonPropertyName("Mode Status DPT", season), modeDpt)
      end
    end
  end

  AddGroupItemsDelayed(10)
end


function SetConnected(bConnected)
  gConnected = gConnected or ""
  if (bConnected ~= gConnected) then
    if (bConnected) then
      MySetMessage("")
    else
      MySetMessage("KNX Offline")
    end
  end
  gConnected = bConnected
end

print("Driver Loaded: " .. os.date())

--[[EXEC_LUA_END]]

gHoldMode = "Off"
if (LuaC4Object.AllowExecute) then C4:AllowExecute(true) end

-- Require OneShotTimer here, since it modifies global OnTimerExpired
require 'OneShotTimer'
